from .dataclasses import (DjangoMigrationConfig)
from .django import (DjangoMigration, DjangoFixtures)
from .funcs import (init_django_app, require_django, log_to_console)
from .tasks import (MsbAppSetupTask, MsbAppPreCommitTask)

__all__ = [
	"init_django_app", "log_to_console", "require_django",
	"DjangoMigration", "DjangoFixtures", "DjangoMigrationConfig",
	"MsbAppSetupTask", "MsbAppPreCommitTask",
]



from .decorators import (validation_schema_wrapper, validate_access, validate_against, validate_request_data)
from .exceptions import *
from .rules import (DefaultRules)
from .schema import (RuleSchema, InputField, ValidationSchema)


class Validate:
	raw_schema = validate_against
	request_data = validate_request_data
	access = validate_access


__all__ = [
	"DefaultRules", "validation_schema_wrapper", "validate_access", "validate_against", "validate_request_data",
	"RuleSchema", "InputField", "Validate", "ValidationSchema",
]



from django.utils.functional import cached_property

from msb_cipher import Cipher
from .constants import TokenConst, AUTH_TOKEN_ACCESS_TYPE_ISR, AUTH_REQUEST_TYPE_ISR
from .session import SessionData


class AuthToken:
	token: dict
	_userid_field: str = TokenConst.userid_field

	def __init__(self, tokendata: dict = None):
		self.token = tokendata if tokendata else dict()

	def __str__(self):
		return f"<{self.__class__.__name__} [{self.id}]: {self.username}>"

	def __hash__(self):
		return hash(self.username)

	def __getattr__(self, attr):
		"""This acts as a backup attribute getter for custom claims defined in Token serializers."""
		return self.token.get(attr, None)

	def set_validation_status(self, status: bool = False):
		self.token['is_valid'] = status

	@property
	def owner(self):
		return Cipher.decrypt(self.token.get('owner'))

	@cached_property
	def id(self):
		return self.token.get(self._userid_field)

	@cached_property
	def userid(self):
		return self.id

	@property
	def is_valid(self):
		return self.token.get('is_valid', False)

	@cached_property
	def session(self) -> SessionData:
		session_data = self.token.get('session', {})
		return session_data if isinstance(session_data, SessionData) else SessionData(**session_data)

	@cached_property
	def access_type(self):
		return Cipher.decrypt(self.token.get('access_type'))

	@property
	def env(self):
		return Cipher.decrypt(self.token.get('env'))

	@cached_property
	def auth_type(self) -> str:
		return self.session.get('auth_type', '')

	@property
	def is_intra_service_requester(self):
		return (
			self.access_type.casefold() == AUTH_TOKEN_ACCESS_TYPE_ISR.casefold()
			and self.auth_type.casefold() == AUTH_REQUEST_TYPE_ISR.casefold()
		)



from rest_framework.permissions import BasePermission


class MsbPermisson(BasePermission):

	def has_permission(self, request, view):
		return True

	def has_object_permission(self, request, view, obj):
		return True


class LoginRequiredPermission(MsbPermisson):

	def has_permission(self, request, view):
		return bool(request.user and request.user.is_authenticated)


class AdminUserPermission(MsbPermisson):
	def has_permission(self, request, view):
		return bool(request.user and request.user.is_staff)


class IntraServiceRequestPermission(LoginRequiredPermission):

	def has_permission(self, request, view) -> bool:
		if super().has_permission(request, view):
			return request.user.is_intra_service_requester
		return False



import os
import sys

from msb_auth.defaults import DefaultJwtAuthSettings, DefaultJwtConfig
from msb_env import (Config, PathConst, EnvDefaults)
from msb_logging import LogConfig

sys.path.extend(PathConst.SYS_PATH_LIST)

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = PathConst.BASE_DIR_PATH

# load configuration from env file
Config.load(env_path=BASE_DIR.__str__(), main_file='.env', local_file='.env.local', )

# Enviroment name
ENVIRONMENT = Config.get('ENVIROMENT').as_str()

# directory where logs will be created
LOGS_DIR = PathConst.LOGS_DIR_PATH

# path where resources are stored
RESOURCE_DIR = PathConst.RESOURCE_DIR_PATH

# path where application will write
WRITABLE_DIR = PathConst.WRITABLE_DIR_PATH

# path where resources are stored
UNITTEST_DATA_DIR = PathConst.TEST_DATA_DIR_PATH

# path where fixtures are stored
FIXTURE_DIRS = PathConst.FIXTURE_DIRS_LIST

# IS DEVELOPEMENT ENVIRONEMR
IS_LOCAL_ENVIRONMENT = Config.is_local_env()

# initailze the default logging configuration
LogConfig.init_default_config(LOGS_DIR, emulate_prod=True)

"""
CORE DJANGO SETTINGS
"""

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = Config.get("SECRET_KEY").as_str(default='')

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = Config.get('DEBUG').as_bool(default=False)

ALLOWED_HOSTS = Config.get('ALLOWED_HOSTS').as_list(default=['*'])

"""
BASE PATH & URL CONFIGURATIONS
"""
# allow admin url
ALLOW_ADMIN_URL = Config.get('ALLOW_ADMIN_URL').as_bool(default=False)

# path to access the admin interface
ADMIN_URL_PATH = Config.get('ADMIN_URL_PATH').as_str(default=None)

APPEND_SLASH = Config.get('APPEND_SLASH').as_bool(default=False)

# common application url
BASE_URL = Config.get('BASE_URL').as_str(default='/')

# installed apps
INSTALLED_APPS = [
	'django.contrib.staticfiles',
	'django.contrib.auth',
	'django.contrib.contenttypes',
	'corsheaders',
	'rest_framework',
	'rest_framework_simplejwt',
]

# Application definition
MIDDLEWARE = [
	'django.middleware.security.SecurityMiddleware',
	'django.contrib.sessions.middleware.SessionMiddleware',
	'corsheaders.middleware.CorsMiddleware',
	'django.middleware.common.CommonMiddleware',
	'django.middleware.csrf.CsrfViewMiddleware',
	'django.contrib.auth.middleware.AuthenticationMiddleware',
	'django.contrib.messages.middleware.MessageMiddleware',
	'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

# Templates Configurations
TEMPLATES = [
	{
		'BACKEND': 'django.template.backends.django.DjangoTemplates',
		'DIRS': [],
		'APP_DIRS': True,
		'OPTIONS': {
			'context_processors': [
				'django.template.context_processors.debug',
				'django.template.context_processors.request',
				'django.contrib.auth.context_processors.auth',
				'django.contrib.messages.context_processors.messages',
			],
		},
	},
]

# Password validation
# https://docs.djangoproject.com/en/4.0/ref/settings/#auth-password-validators
AUTH_PASSWORD_VALIDATORS = [
	{
		'NAME': 'django.contrib.backends.password_validation.UserAttributeSimilarityValidator',
	},
	{
		'NAME': 'django.contrib.backends.password_validation.MinimumLengthValidator',
	},
	{
		'NAME': 'django.contrib.backends.password_validation.CommonPasswordValidator',
	},
	{
		'NAME': 'django.contrib.backends.password_validation.NumericPasswordValidator',
	},
]

ROOT_URLCONF = 'urls'

WSGI_APPLICATION = 'app.wsgi.application'

# Default primary key field type
# https://docs.djangoproject.com/en/4.0/ref/settings/#default-auto-field
DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.0/howto/static-files/

STATIC_URL = 'static/'

STATIC_ROOT = os.path.join(BASE_DIR, 'static')

# Internationalization
# https://docs.djangoproject.com/en/4.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_TZ = False

"""
SECURITY CONFIGURATIONS
"""

# Secure csrf cookie
CSRF_COOKIE_SECURE = EnvDefaults.CSRF_COOKIE_SECURE

# Secure session cookie
SESSION_COOKIE_SECURE = EnvDefaults.SESSION_COOKIE_SECURE

# make csrf cokkied http only
CSRF_COOKIE_HTTPONLY = EnvDefaults.CSRF_COOKIE_HTTPONLY

# make csrf cokkied http only
SESSION_COOKIE_HTTPONLY = EnvDefaults.SESSION_COOKIE_HTTPONLY

# secure the ssl redirect
SECURE_SSL_REDIRECT = Config.get('SECURE_SSL_REDIRECT').as_bool(default=False)

# domain for which the session cookie is valid
SESSION_COOKIE_DOMAIN = Config.get('SESSION_COOKIE_DOMAIN').as_str(default='127.0.0.1')

# allow all origins to make requests
CORS_ALLOW_ALL_ORIGINS = Config.get('CORS_ALLOW_ALL_ORIGINS').as_bool(default=True)

# allowed origins
CORS_ALLOWED_ORIGINS = Config.get('CORS_ALLOWED_ORIGINS').as_list()

CORS_ALLOW_METHODS = EnvDefaults.CORS_ALLOW_METHODS

CORS_ALLOW_HEADERS = EnvDefaults.CORS_ALLOW_HEADERS

CORS_PREFLIGHT_MAX_AGE = EnvDefaults.CORS_PREFLIGHT_MAX_AGE

CORS_ALLOW_CREDENTIALS = EnvDefaults.CORS_ALLOW_CREDENTIALS

ALLOW_API_DOCUMENTATION = Config.get("ALLOW_API_DOCUMENTATION").as_bool(default=False)

"""
DJANGO REST FRAMEWORK CONFIGURATIONS
:see https://www.django-rest-framework.org/api-guide/settings/ for details.
"""
REST_FRAMEWORK = dict()
REST_FRAMEWORK["DEFAULT_PAGINATION_CLASS"] = "rest_framework.pagination.PageNumberPagination"
REST_FRAMEWORK["PAGE_SIZE"] = 10
REST_FRAMEWORK["DEFAULT_PERMISSION_CLASSES"] = ('rest_framework.permissions.IsAuthenticated',)
REST_FRAMEWORK["DEFAULT_RENDERER_CLASSES"] = ["rest_framework.renderers.JSONRenderer", ]
REST_FRAMEWORK["TEST_REQUEST_DEFAULT_FORMAT"] = 'json'

if Config.is_local_env():
	REST_FRAMEWORK['DEFAULT_RENDERER_CLASSES'] += ["rest_framework.renderers.BrowsableAPIRenderer"]

JWT_AUTH_CONFIG = DefaultJwtAuthSettings(
	signing_key=Config.get("JWT_TOKEN_SIGNING_KEY").as_str(default=''),
	verifying_key=Config.get("JWT_TOKEN_VERIFY_KEY").as_str(default=''),
	access_token_lifetime=Config.get("JWT_ACCESS_TOKEN_VALIDITY").as_int(default=DefaultJwtConfig.access_token_lifetime),
	refresh_token_lifetime=Config.get("JWT_ACCESS_TOKEN_VALIDITY").as_int(default=DefaultJwtConfig.refresh_token_lifetime),
)

# default classes that will authenticate all requests
REST_FRAMEWORK["DEFAULT_AUTHENTICATION_CLASSES"] = JWT_AUTH_CONFIG.authentication_classes

# load meme-type configuration
import mimetypes

mimetypes.add_type("text/css", ".css", True)

"""
CELERY CONFIGURATIONS
"""
# Message Broker settings.
CELERY_BROKER_URL = Config.get('CELERY_BROKER_URL').as_str()
# List of modules to import when celery starts.
CELERY_IMPORTS = []
# Using the database to store task state and results.
CELERY_RESULT_BACKEND = Config.get('CELERY_RESULT_BACKEND').as_str()
CELERY_RESULT_EXTENDED = True
# Format in which the tasks are stored
CELERY_ACCEPT_CONTENT = ["json"]
CELERY_TASK_SERIALIZER = "json"

"""
MSB CONFIGURATIONS
"""
MSB_DB_ENCRYPT_PRIVATE_FIELDS = EnvDefaults.MSB_DB_ENCRYPT_PRIVATE_FIELDS
MSB_DB_PK_IS_PRIVATE_FIELD = EnvDefaults.MSB_DB_PK_IS_PRIVATE_FIELD
MSB_JWT_TRUSTED_OWNERS = EnvDefaults.MSB_JWT_TRUSTED_OWNERS
MSB_JWT_VERIFY_OWNER = EnvDefaults.MSB_JWT_VERIFY_OWNER
MSB_JWT_VERIFY_SUBSCRIPTIONS = EnvDefaults.MSB_JWT_VERIFY_SUBSCRIPTIONS
MSB_JWT_VERIFY_ENVIRONMENT = EnvDefaults.MSB_JWT_VERIFY_ENVIRONMENT



from django.urls import include, path

from .constants import CRUD_URL_PK_NAME, CrudMethods
from .constants import REQUEST_METHODS as HttpVerbs


class MsbApiDefaultRoutes:

	@staticmethod
	def search_action_routes(cls, search: bool):
		actions = {HttpVerbs.POST: (CrudMethods.search if search == True else CrudMethods.not_found)}
		return path("/search", cls.as_view(actions=actions))

	@staticmethod
	def single_action_routes(cls, retrieve: bool = True, update: bool = True, delete: bool = True):
		actions = {
			HttpVerbs.GET: (CrudMethods.retrieve if retrieve == True else CrudMethods.not_found),
			HttpVerbs.PUT: (CrudMethods.update if update == True else CrudMethods.not_found),
			HttpVerbs.DELETE: (CrudMethods.delete if delete == True else CrudMethods.not_found),
		}
		return path(f"/<str:{(cls.pk_name or CRUD_URL_PK_NAME)}>", cls.as_view(actions=actions))

	@staticmethod
	def bulk_action_routes(cls, create: bool = True, list: bool = True, bulk_update: bool = True, bulk_delete: bool = True):
		actions = {
			HttpVerbs.GET: (CrudMethods.list if list == True else CrudMethods.not_found),
			HttpVerbs.POST: (CrudMethods.create if create == True else CrudMethods.not_found),
			HttpVerbs.PUT: (CrudMethods.bulk_update if bulk_update == True else CrudMethods.not_found),
			HttpVerbs.DELETE: (CrudMethods.bulk_delete if bulk_delete == True else CrudMethods.not_found),
		}
		return path("", cls.as_view(actions=actions))


class MsbApiRoutes:

	@classmethod
	def crud_routes(cls, create: bool = True, retrieve: bool = True, update: bool = True, delete: bool = True,
	                list: bool = True, bulk_update: bool = True, bulk_delete: bool = True, search: bool = True):
		return include([
			MsbApiDefaultRoutes.search_action_routes(cls, search),
			MsbApiDefaultRoutes.bulk_action_routes(cls, create, list, bulk_update, bulk_delete),
			MsbApiDefaultRoutes.single_action_routes(cls, retrieve, update, delete),
		])



import os
from pathlib import Path

from pdf2docx import Converter

from .core import FileExceptions
from .core import FileInterface, FileBase
from .core import FileMessage
from .core import GeneratedFile
from .pdf import PdfFile


class DocxFile(FileBase, FileInterface):
	_extension = None

	def __init__(self):
		self._extension = ".docx"

	def save(self, filename: str, location: str) -> GeneratedFile:
		file = GeneratedFile()
		try:
			content = self.content
			doc_filepath = os.path.join(location, filename + self._extension)
			Path(location).mkdir(parents=True, exist_ok=True)
			if os.path.exists(doc_filepath) and self.file_overwrite is False:
				raise FileExceptions.DocxExistsException
			"""Generate PDF"""
			pdf = PdfFile().set_content(content)
			if self.file_overwrite is True:
				pdf.set_file_overwrite()
			generated_pdf = pdf.save(location=location, filename=filename)
			pdf_path = generated_pdf.path

			if pdf_path != "":
				"""Generate docx file"""
				cv = Converter(pdf_path)
				cv.convert(doc_filepath)
				cv.close()
				if not cv:
					FileExceptions.DocxException()
				else:
					file.path = doc_filepath
					file.message = FileMessage.doc()
					os.remove(pdf_path)
			else:
				raise FileExceptions.PdfException()
		except Exception as e:
			file.message = e
			file.error = True
		finally:
			return file



import logging
import os
from datetime import datetime

from msb_env import (Config)
from .constants import (
	LOG_DATETIME_FORMAT, SIMPLE_LOG_FORMAT,
	DEFAULT_FILE_HANDLER_NAME, DEFAULT_CONSOLE_HANDLER_NAME
)


def make_log_file_name(dir: str = '', filename: str = None):
	filename_prefix = f"{filename}-" if filename not in ['', None] else ''
	return os.path.join(dir, f"{filename_prefix}{datetime.today().strftime('%Y-%m-%d')}.log")


class LogConfig:
	_default_log_level: str
	_default_format_str: str

	def __init__(self, **kwargs):
		self._default_log_level = kwargs.get('default_level') or logging.WARNING
		self._default_format_str = None

	@staticmethod
	def init_default_config(logsdir: str, default_level: str = logging.INFO, emulate_prod=False):

		default_log_config = LogConfig(default_level=default_level)

		console_log_handler = default_log_config.get_console_handler(
			DEFAULT_CONSOLE_HANDLER_NAME,
			level=logging.INFO,
		)

		system_exceptions_file_handler = default_log_config.get_file_handler(
			DEFAULT_FILE_HANDLER_NAME, logs_dir=logsdir, level=logging.ERROR,
			formatter=SIMPLE_LOG_FORMAT, filename='exceptions'
		)

		debug_file_log_handler = default_log_config.get_file_handler(
			DEFAULT_FILE_HANDLER_NAME, logs_dir=logsdir, level=logging.DEBUG,
			formatter=SIMPLE_LOG_FORMAT, filename='log'
		)

		_root_logging_handlers = [console_log_handler]
		_debug_logging_handlers = [debug_file_log_handler]

		if Config.is_local_env():
			_root_logging_handlers.append(console_log_handler)
			_debug_logging_handlers.append(console_log_handler)
			if emulate_prod:
				_root_logging_handlers.append(system_exceptions_file_handler)
				_debug_logging_handlers.append(system_exceptions_file_handler)

		else:
			_root_logging_handlers.append(system_exceptions_file_handler)

		default_log_config.add_logger('console', handlers=[console_log_handler])
		default_log_config.add_logger('root', handlers=_root_logging_handlers)
		default_log_config.add_logger('debug', handlers=_debug_logging_handlers, level=logging.INFO)

	def add_formatter(self, name: str, format: str = ''):
		self.formatters[name] = {"format": format}

	def add_filters(self, name: str, handler_class: str):
		self.filters[name] = {
			'()': handler_class,
		}

	def add_logger(self, name: str, level: str = None, handlers: list = None, **kwargs):
		try:
			log_handlers = handlers if isinstance(handlers, list) else []
			_logger = logging.getLogger(name=name)
			log_level = level if level is not None else self._default_log_level

			_logger.setLevel(level=log_level)
			_logger.propagate = kwargs.get('propagate') == True

			for log_handler in log_handlers:
				_logger.addHandler(log_handler)

		except Exception as e:
			print(f"Logger registration failed for '{name}' :\n{e}")

	def get_file_handler(self, name, logs_dir: str, filename: str = None, handler_class=None, **kwargs):
		from logging.handlers import TimedRotatingFileHandler
		handler_class = handler_class if callable(handler_class) else TimedRotatingFileHandler
		init_kwargs = dict(filename=make_log_file_name(dir=logs_dir, filename=filename), when='midnight')
		return self.__build_log_handler(
			name=name, log_handler=handler_class, init_kwargs=init_kwargs, level=kwargs.get('level'),
			formatter=kwargs.get("formatter"), filters=kwargs.get('filters')
		)

	def get_email_handler(self, name, email_backend, handler_class: str = None, **kwargs):
		from django.utils.log import AdminEmailHandler
		handler_class = handler_class if callable(handler_class) else AdminEmailHandler
		init_kwargs = dict(
			include_html=kwargs.get('include_html') == True,
			email_backend=email_backend,
			reporter_class=kwargs.get('reporter_class')
		)
		return self.__build_log_handler(
			name=name, log_handler=handler_class, init_kwargs=init_kwargs, level=kwargs.get('level'),
			formatter=kwargs.get("formatter"), filters=kwargs.get('filters')
		)

	def get_console_handler(self, name, handler_class: str = None, **kwargs):
		handler_class = handler_class if callable(handler_class) else logging.StreamHandler
		init_kwargs = {}
		return self.__build_log_handler(
			name=name, log_handler=handler_class, init_kwargs=init_kwargs,
			level=kwargs.get('level'), formatter=kwargs.get("formatter"), filters=kwargs.get('filters')
		)

	def get_database_handler(self, name, func, **kwargs):
		from .handlers import DatabaseLogger
		init_kwargs = dict(callback_func=func)
		return self.__build_log_handler(
			name=name, log_handler=DatabaseLogger, init_kwargs=init_kwargs,
			level=kwargs.get('level'), formatter=kwargs.get("formatter"), filters=kwargs.get('filters')
		)

	def __build_log_handler(self, log_handler: logging.Handler, init_kwargs: dict = None, **kwargs):
		try:
			init_kwargs = init_kwargs if isinstance(init_kwargs, dict) else {}
			log_formatter = kwargs.get("formatter") or SIMPLE_LOG_FORMAT
			log_level = kwargs.get("level") or logging.WARNING
			_handler = log_handler(**init_kwargs)
			_handler.setLevel(level=log_level)
			_handler.setFormatter(logging.Formatter(log_formatter, datefmt=LOG_DATETIME_FORMAT))
			return _handler
		except Exception as e:
			logging.warning(e)



from msb_apis import (
	ApiView, IntraServiceRequestApi, permissions, api_details, TokenUser, SessionData, ApiResponse,
	RequestInfo, RequestHeaders, ApiViewset, RestResponse
)
from msb_auth.decorators import (management_api, require_permission, require_authentication, )
from msb_cipher import (Cipher)
from msb_dataclasses import (SearchParameter, SearchParameterRequest, Singleton)
from msb_db.models import (MsbModel, MsbModelMetaFields, MsbModelManager, model_fields)
from msb_exceptions import (ApiException, AppException, CrudApiException)
from msb_services import (ApiService, ApiServiceExceptions, CrudApiException, IntraServiceRequestService)
from msb_validation import (validate_request_data, DefaultRules, ValidationSchema, InputField)

__all__ = [

	# core msb_apis imports
	"ApiView", "IntraServiceRequestApi", "ApiViewset", "api_details", "ApiService", "IntraServiceRequestService",
	"permissions", "management_api", "validate_request_data", "require_permission", "require_authentication",
	"TokenUser", "SessionData", "ApiResponse", "RequestInfo", "RequestHeaders", "RestResponse",

	# core msb_exceptions imports
	"ApiException", "CrudApiException", "AppException", "ApiServiceExceptions",

	# core msb_validation imports
	"DefaultRules", "ValidationSchema", "InputField",

	# core msb_db imports
	"MsbModel", "MsbModelMetaFields", "MsbModelManager", "model_fields",

	# core msb_dataclasses imports
	"SearchParameter", "SearchParameterRequest", "Singleton",

	# others
	"Cipher"
]



import logging
from dataclasses import dataclass

from msb_env import NameConst


@dataclass
class DbVendorConfig:
	query_to_list_tables: str
	query_to_drop_table: str
	excluded_tables: list

	def queries_to_drop_multiple_tables(self, table_list) -> list:
		_drop_table_queries = []
		if isinstance(table_list, list) and len(table_list) > 0:
			for table in table_list:
				if table[0] not in self.excluded_tables and self.query_to_drop_table.strip(" ") != "":
					_drop_table_queries.append(self.query_to_drop_table.format(table_name=table[0]))
		return _drop_table_queries


class DjangoMigrationConfig:
	_is_local_env: bool = False
	_remove_migration_files: bool = False
	_drop_tables_from_db: bool = False
	_db_to_drop_tables_from: list = list()
	_apps_to_migrate: list = list()
	_dbs_to_migrate: list = list()

	@property
	def is_local_env(self) -> bool:
		return self._is_local_env

	@property
	def remove_migration_files(self) -> bool:
		return self._remove_migration_files

	@property
	def drop_tables_from_db(self) -> bool:
		return self._drop_tables_from_db

	@property
	def db_to_drop_tables_from(self) -> tuple:
		return self._db_to_drop_tables_from

	@property
	def apps_to_migrate(self) -> tuple:
		return self._apps_to_migrate

	@property
	def dbs_to_migrate(self) -> tuple:
		return self._dbs_to_migrate

	def set_apps_to_migrate(self, *app_list):
		self._apps_to_migrate = app_list

	def set_dbs_to_migrate(self, *db_list):
		self._dbs_to_migrate = db_list

	def __init__(self, env: str, **kwargs):

		if isinstance(env, str):

			self._is_local_env = env.lower() == NameConst.LOCAL_ENV_NAME

			if self.is_local_env:
				self._remove_migration_files = kwargs.get("remove_migration_files") == True
				self._drop_tables_from_db = kwargs.get("drop_tables_from_db") == True
				self._db_to_drop_tables_from = kwargs.get("db_to_drop_tables_from") or [NameConst.DEFAULT_DATABASE_NAME]
		else:
			logging.warning("Invalid or empty environment name.")


class DjangoFixtureDirs:
	pass



from functools import wraps

from rest_framework import request as drf_request

from msb_http import ApiResponse
from .exceptions import (ValidatorExceptions, InvalidPayloadFormat, InvalidPayloadException, InvalidParamsException)
from .schema import ValidationSchema

BULK_SCHEMA_NAME = 'schema_list'


def __parse_payload_from_request(request: drf_request.Request):
	try:
		_payload = request.data
		return _payload
	except Exception as e:
		raise InvalidPayloadFormat


def __parse_schema_and_inp_for_bulk_validation(schema: ValidationSchema, inp, bulk=False):
	_parsed_schema = schema.parsed_schema
	_parsed_inp = inp
	if bulk:
		_parsed_schema = {BULK_SCHEMA_NAME: {'type': 'list', 'schema': {'type': 'dict', 'schema': _parsed_schema}}}
		_parsed_inp = {BULK_SCHEMA_NAME: inp}

	return _parsed_schema, _parsed_inp


def __format_list_errors_to_string(errors: dict):
	return {k: v[0] if isinstance(v, list) and len(v) > 0 else v for k, v in errors.items()}


def __formatted_validation_errors(errors: dict, bulk: bool = False):
	if isinstance(errors, dict) and len(errors.keys()) > 0:
		try:
			if bulk:
				error_list: dict = errors.get(BULK_SCHEMA_NAME)[0]
				for error_index, errors in error_list.items():
					error_list.update({error_index: __format_list_errors_to_string(errors=errors[0])})
				return error_list
			else:
				return __format_list_errors_to_string(errors=errors)

		except Exception:
			raise ValidatorExceptions.ErrorFormattingFailed
	return None


def validation_schema_wrapper(klass):
	def get(key: str, default=None) -> ValidationSchema | None:
		if hasattr(klass, key) and isinstance((schema := getattr(klass, key)), ValidationSchema):
			return schema
		return default

	setattr(klass, "get", get)
	return klass


def validate_against(schema: ValidationSchema = None, inp=None, unknown=False, bulk: bool = False):
	if not isinstance(schema, ValidationSchema):
		raise ValidatorExceptions.InvalidValidationSchema

	if not ((bulk and isinstance(inp, list)) or (not bulk and isinstance(inp, dict))):
		raise ValidatorExceptions.InvalidValidationInput

	errors = {}
	schema, inp = __parse_schema_and_inp_for_bulk_validation(schema, inp, bulk=bulk)
	try:
		from cerberus.validator import Validator
		_validator = Validator(schema, allow_unknown=unknown)
		_validator.validate(inp)
		errors = _validator.errors
	except Exception as e:
		raise ValidatorExceptions.ValidationFailed
	return __formatted_validation_errors(errors, bulk=bulk)


def validate_request_data(payload_schema: ValidationSchema = None, param_schema: ValidationSchema = None, **opt):
	allow_unknown = opt.get("unknown") == True
	bulk_inp = opt.get("bulk_inp") == True

	def outer_func(_func):
		@wraps(_func)
		def inner_func(cls, request: drf_request.Request, *args, **kwargs):

			try:
				is_valid_payload_schema = isinstance(payload_schema, ValidationSchema)
				is_valid_param_schema = isinstance(param_schema, ValidationSchema)
				if (not is_valid_payload_schema and not is_valid_param_schema):
					raise ValidatorExceptions.InvalidValidationSchema

				# validate paload data if payload rules are defined
				if isinstance(payload_schema, ValidationSchema):
					_payload = __parse_payload_from_request(request=request)
					payload_validation_errors = validate_against(
						schema=payload_schema, inp=_payload, unknown=allow_unknown, bulk=bulk_inp
					)
					if payload_validation_errors:
						raise InvalidPayloadException(errors=payload_validation_errors)

				# validate parameter data like from URL/Query string if parameter
				# validation rules are defined
				if isinstance(param_schema, ValidationSchema):
					params_validation_errors = validate_against(schema=param_schema, inp=kwargs, unknown=True)
					if params_validation_errors:
						raise InvalidParamsException(errors=params_validation_errors)

				return _func(cls, *args, **dict(request=request, **kwargs))

			except InvalidPayloadException as e:
				return ApiResponse.exception(e)

			except InvalidParamsException as e:
				return ApiResponse.exception(e)

			except Exception as e:
				return ApiResponse.exception(e)

		return inner_func

	return outer_func


def validate_access(group: list = None, action: str = '', **opt):
	def decorator_function(_decorated_function):
		@wraps(_decorated_function)
		def wrapper_function(cls, request: drf_request.Request, *args, **kwargs):
			try:
				return _decorated_function(cls, *args, **dict(request=request, **kwargs))
			except Exception as e:
				ApiResponse.error(e)

		return wrapper_function

	return decorator_function



from __future__ import annotations

from rest_framework import status
from rest_framework.response import Response as RestResponse

from msb_exceptions import AppException
from msb_exceptions import CustomExceptionHandler


class ApiResponse:
	data_types = [list, dict]

	@staticmethod
	def _dispatch(data: dict = None, status=status.HTTP_200_OK) -> RestResponse:
		headers = dict()
		_response = RestResponse(data=data, headers=headers, status=status)
		return _response

	@staticmethod
	def success(data=None, msg: str = 'ok') -> RestResponse:
		response_data = dict(
			success=True, message=msg,
			code=200, data=([] if type(data) not in ApiResponse.data_types else data)
		)
		return ApiResponse._dispatch(data=response_data)

	@staticmethod
	def error(msg: str = "Failure", err_code=600, data=None, http_code=status.HTTP_200_OK) -> RestResponse:
		response_data = dict(
			success=False,
			message=msg,
			code=err_code,
			data=([] if type(data) not in ApiResponse.data_types else data)
		)
		return ApiResponse._dispatch(data=response_data, status=http_code)

	@staticmethod
	def exception(e: AppException | Exception = None, fallback: bool = True) -> RestResponse:
		CustomExceptionHandler.handle_exception(e=e)
		if not isinstance(e, AppException):
			return ApiResponse.internal_server_error()

		return ApiResponse.error(
			msg=e.message if hasattr(e, 'message') else str(e),
			data=e.errors if hasattr(e, 'errors') else [],
			err_code=e.code if hasattr(e, 'code') else 500
		)

	@staticmethod
	def not_found(msg: str = "Resource Not Found.", **kwargs):
		response_data = dict(success=False, message=msg, code=404, data=[])
		return ApiResponse._dispatch(data=response_data, status=status.HTTP_404_NOT_FOUND)

	@staticmethod
	def authentication_failed(msg: str = "Authentication Failed.", **kwargs):
		response_data = dict(success=False, message=msg, code=401, data=[])
		return ApiResponse._dispatch(data=response_data, status=status.HTTP_401_UNAUTHORIZED)

	@staticmethod
	def forbidden(msg: str = "Access Forbidden.", **kwargs):
		response_data = dict(success=False, message=msg, code=403, data=[])
		return ApiResponse._dispatch(data=response_data, status=status.HTTP_403_FORBIDDEN)

	@staticmethod
	def bad_request(msg: str = "Bad Request.", **kwargs):
		response_data = dict(success=False, message=msg, code=400, data=[])
		return ApiResponse._dispatch(data=response_data, status=status.HTTP_400_BAD_REQUEST)

	@staticmethod
	def method_not_allowed(msg: str = "Method Not Allowed.", **kwargs):
		response_data = dict(success=False, message=msg, code=405, data=[])
		return ApiResponse._dispatch(data=response_data, status=status.HTTP_405_METHOD_NOT_ALLOWED)

	@staticmethod
	def internal_server_error(msg: str = "Internal Server Error", **kwargs):
		response_data = dict(success=False, message=msg, code=500, data=[])
		return ApiResponse._dispatch(
			data=response_data,
			status=status.HTTP_500_INTERNAL_SERVER_ERROR
		)

	@staticmethod
	def formatted_error_response(response: RestResponse) -> RestResponse:
		_code = response.status_code or None
		_message = response.data.get('detail') or None
		if _code is not None and _message is not None:
			return ApiResponse.error(err_code=_code, msg=_message.replace('"', "'"), http_code=_code)
		return ApiResponse.internal_server_error()



import os


class EmailConfigData(object):
	_smtp_host: str = 'localhost'
	_smtp_port: int = 25
	_smtp_user: str = ''
	_smtp_password: str = ''
	_smtp_use_tls: bool = False
	_smtp_use_ssl: bool = False
	_smtp_timeout: int = 0
	_smtp_ssl_keyfile: str = ''
	_smtp_ssl_certfile: str = ''

	def __init__(self, namespace=None):
		namespace = str(namespace).upper()
		self._smtp_host = os.environ.get(f'{namespace}_SMTP_HOST', default='')
		self._smtp_port = os.environ.get(f'{namespace}_SMTP_PORT', default='')
		self._smtp_username = os.environ.get(f'{namespace}_SMTP_USER', default='')
		self._smtp_password = os.environ.get(f'{namespace}_SMTP_PASSWORD', default='')
		self._smtp_use_tls = (int(os.environ.get(f'{namespace}_SMTP_USE_TLS', default=0)) == 1)
		self._smtp_use_ssl = (int(os.environ.get(f'{namespace}_SMTP_USE_SMTP', default=0)) == 1)
		self._smtp_timeout = int(os.environ.get(f'{namespace}_SMTP_TIMEOUT', default=30))
		self._smtp_ssl_keyfile = os.environ.get(f'{namespace}_SMTP_SSL_KEY_FILE', default=None)
		self._smtp_ssl_certfile = os.environ.get(f'{namespace}_SMTP_SSL_CERTIFICATE', default=None)

	def __str__(self):
		return str(self.__dict__)

	@property
	def host(self):
		return self._smtp_host

	@property
	def port(self):
		return self._smtp_port

	@property
	def user(self):
		return self._smtp_user

	@property
	def password(self):
		return self._smtp_password

	@property
	def timeout(self):
		return self._smtp_timeout

	@property
	def use_ssl(self):
		return self._smtp_use_ssl

	@property
	def use_tls(self):
		return self._smtp_use_tls

	@property
	def ssl_key(self):
		return self._smtp_ssl_keyfile

	@property
	def ssl_cert(self):
		return self._smtp_ssl_certfile


class EmailConfig():
	__config = dict()

	def __init__(self, *namespace_list):
		for namespace in namespace_list:
			self.__config[str(namespace).lower()] = EmailConfigData(namespace=namespace)

	def get(self, namespace: str = 'default') -> EmailConfigData:
		return self.__config.get(namespace.lower())



from msb_exceptions import AppException


class ValidatorExceptions:
	class InvalidValidationSchema(AppException):
		_message = 'Validation schema should be of type ValidationSchema.'

	class InvalidValidationInput(AppException):
		_message = 'Validation input should be of type dict or list.'

	class SchemaValidationFailed(AppException):
		_message = 'Failed to validate the input against the given schema.'

	class ErrorFormattingFailed(AppException):
		_message = 'Failed to format the validation errors.'

	class ValidationFailed(AppException):
		_message = 'Failed to validate the request data.'


class InvalidPayloadException(AppException):
	_message = 'Invalid payload recieved.'


class InvalidParamsException(AppException):
	_message = 'Invalid request parameters recieved.'


class InvalidPayloadFormat(AppException):
	_message = 'Malformed request data recieved.'


__all__ = ["ValidatorExceptions", "InvalidParamsException", "InvalidPayloadException", "InvalidPayloadFormat"]



import logging

from django.db.models import (manager, DateTimeField, IntegerField, CharField, TextField)

from msb_logging import (
	LOG_DB, LOG_TABLE, CHOICES_LOG_LEVELS, DEFAULT_DBLOG_COLUMN_MAPPINGS,
	SYSTEM_DBLOG_COLUMN_MAPPINGS
)
from .models import (MsbModel)


class LoggingModelManager(manager.Manager):
	log_db = LOG_DB

	def get_queryset(self):
		return super(LoggingModelManager, self).get_queryset().using(self.log_db)

	@property
	def exceptions(self):
		return self.get_queryset().filter(level=logging.CRITICAL)

	@property
	def errors(self):
		return self.get_queryset().filter(level=logging.ERROR)

	@property
	def warnings(self):
		return self.get_queryset().filter(level=logging.WARNING)

	@property
	def info(self):
		return self.get_queryset().filter(level=logging.INFO)


class LoggingModel(MsbModel):
	_log_table_columns_mapping: dict = DEFAULT_DBLOG_COLUMN_MAPPINGS

	class Meta:
		abstract = True
		managed = False

	level = CharField(choices=CHOICES_LOG_LEVELS, max_length=255, default=logging.INFO)
	message = TextField(default=None, null=True)
	created_at = DateTimeField(auto_now_add=True, db_column='created_at')

	objects = LoggingModelManager()

	@classmethod
	def create(cls, *args, **kwargs):

		init_kwargs = kwargs
		init_args = args
		try:
			if isinstance(cls._log_table_columns_mapping, dict) and len(cls._log_table_columns_mapping.keys()) > 0:
				init_kwargs = {k: kwargs.get(v) for k, v in cls._log_table_columns_mapping.items()}

			return cls(*init_args, **init_kwargs).save()

		except Exception as e:
			return logging.warning(e)


class SystemLogModel(LoggingModel):
	_log_table_columns_mapping: dict = SYSTEM_DBLOG_COLUMN_MAPPINGS

	class Meta:
		db_table = LOG_TABLE
		abstract = True
		managed = False

	logger = CharField(max_length=255, default='root')

	msg = TextField(default=None)
	module = CharField(max_length=255, default=None, null=True)
	func_name = CharField(max_length=255, default=None, null=True)
	thread_name = CharField(max_length=255, default=None, null=True)
	process_name = CharField(max_length=255, default=None, null=True)
	filename = CharField(max_length=255, default=None, null=True)

	args = TextField(default=None, null=True)
	pathname = TextField(default=None, null=True)
	exc_info = TextField(default=None, null=True)
	exc_text = TextField(default=None, null=True)
	stack_info = TextField(default=None, null=True)
	lineno = IntegerField(default=0, null=True)

	objects = LoggingModelManager()



from django.db.models import (TextField, BigAutoField)

from msb_cipher import Cipher


class EncryptedDatabaseField(TextField):
	description = "Encrypted Field Holds Private Data"
	_type = str

	def __init__(self, *args, **kwargs):
		kwargs['blank'] = True
		kwargs['null'] = True
		super().__init__(*args, **kwargs)

	def deconstruct(self):
		name, path, args, kwargs = super().deconstruct()
		del kwargs["blank"]
		del kwargs["null"]
		return name, path, args, kwargs

	def get_prep_value(self, value):
		return Cipher.encrypt(data=super().get_prep_value(value))

	def from_db_value(self, value, expression, connection):
		return Cipher.decrypt(data=value)

	def to_python(self, value):
		return self._type(Cipher.decrypt(data=value))


class EncryptedInteger(EncryptedDatabaseField):
	_type = int


class EncryptedString(EncryptedDatabaseField):
	_type = str


class EncryptedFloat(EncryptedDatabaseField):
	_type = float


class EncryptedBool(EncryptedDatabaseField):
	_type = bool


class EncryptedPrimaryKey(BigAutoField):
	def encrypted(self):
		return Cipher.encrypt(super(EncryptedPrimaryKey, self).get_prep_value(''))



from typing import Union

from django.db.models import QuerySet
from rest_framework import (viewsets, serializers, exceptions)

from msb_auth import permissions
from msb_auth.session import SessionData
from msb_auth.users import TokenUser
from msb_http import (ApiResponse, RequestInfo, RequestHeaders)
from .constants import DEFAULT_LOGGER_NAME


def api_details(request=None, ver='', name=''):
	return ApiResponse.success(
		data=dict(method=request.method, version=ver, name=name)
	)


class ApiView(viewsets.GenericViewSet):
	permission_classes = (permissions.LoginRequiredPermission,)
	serializer_class = serializers.Serializer
	default_logger: str = DEFAULT_LOGGER_NAME

	def handle_exception(self, exc):
		"""
		override parent exception handler, to send custom error messages as a response
		"""
		try:
			_auth_exceptions = (exceptions.NotAuthenticated, exceptions.AuthenticationFailed)
			if isinstance(exc, _auth_exceptions):
				if (auth_header := self.get_authenticate_header(self.request)) is not None:
					return self.api_response.authentication_failed()
				else:
					return self.api_response.forbidden()

			return self.api_response.formatted_error_response(super().handle_exception(exc=exc))
		except Exception as e:
			return self.api_response.exception(e)

	@property
	def request_info(self) -> RequestInfo:
		return RequestInfo(meta=self.request.META)

	@property
	def request_headers(self) -> RequestHeaders:
		return RequestHeaders(headers=self.request.headers)

	@property
	def token(self) -> TokenUser:
		return self.request.user

	@property
	def user(self) -> SessionData:
		return self.token.session

	@property
	def payload(self) -> dict | list:
		return self.request.data if type(self.request.data) in [list, dict] else {}

	@property
	def params(self) -> dict:
		return self.request.query_params.dict()

	@property
	def logger(self):
		import logging
		return logging.getLogger(self.default_logger)

	@property
	def api_response(self):
		return ApiResponse

	def serializer(self, data: list | dict | QuerySet = None) -> Union[list, dict]:
		if isinstance(data, dict):
			return data

		return (
			[item.dict() if hasattr(item, 'dict') else item for item in data]
		) if (type(data) in [list, QuerySet]) else []


class IntraServiceRequestApi(ApiView):
	permission_classes = (permissions.LoginRequiredPermission,)



from ._email import (EmailConfig, EmailConfigData)
from ._singleton import (Singleton)
from ._wrappers import (ConfigObject)
from .search_parameter import (SearchParameter, SearchParameterRequest)

__all__ = [
	"Singleton", "SearchParameter", "SearchParameterRequest",
	"EmailConfig", "EmailConfigData", "ConfigObject"
]



import os
from io import BytesIO
from pathlib import Path

from xhtml2pdf import pisa

from .core import FileExceptions
from .core import FileInterface, FileBase
from .core import FileMessage
from .core import GeneratedFile


class PdfFile(FileBase, FileInterface):
	_extension = None

	def __init__(self):
		self._extension = ".pdf"

	def save(self, filename: str, location: str) -> GeneratedFile:
		file = GeneratedFile()
		try:
			filepath = os.path.join(location, filename + self._extension)
			if os.path.exists(filepath) and self.file_overwrite is False:
				raise FileExceptions.PdfExistsException
			Path(location).mkdir(parents=True, exist_ok=True)
			with open(filepath, 'wb+') as output:
				content = self.content
				pdf = pisa.pisaDocument(BytesIO(content.encode("ISO-8859-1")), output)
				if not pdf.err:
					file.path = filepath
					file.message = FileMessage.pdf()
				else:
					raise FileExceptions.PdfException
		except Exception as e:
			file.message = e
			file.error = True
		finally:
			return file



from msb_env import NameConst, PathConst
from .dataclasses import DbVendorConfig


def log_to_console(msg, format=False, sep="\n"):
	_log_message = f"\n\n{f'[ {msg} ] ' :*^100}" if format else f"{msg}"
	return print(_log_message, sep=sep)


def get_django_db_vendor_config(db_connection) -> DbVendorConfig:
	from .constants import DJANGO_MIGRATION_DB_VENDOR_CONFIG
	if db_connection and hasattr(db_connection, 'vendor'):
		return DJANGO_MIGRATION_DB_VENDOR_CONFIG.get(getattr(db_connection, 'vendor'))
	return None


def init_django_app(settings_dir: str = NameConst.APP_DIR_NAME, *sys_pathlist, **kwargs):
	import django, sys, os
	sys.path.extend([*PathConst.SYS_PATH_LIST, *sys_pathlist])
	os.environ.setdefault("DJANGO_SETTINGS_MODULE", f"{settings_dir}.settings")
	os.environ.setdefault("PYTHONUNBUFFERED", "1")

	django.setup()


def require_django(_func):
	def inner_func(*args, **kwargs):
		init_django_app()
		return _func(*args, **kwargs)

	return inner_func



from .constants import *
from .wrappers import (
	create_datetime, parse_datetime, current_time, timezone_offset,
	current_timestamp, current_date, timestamp_diff, convert_to_utc,
)

__all__ = [
	"create_datetime", "parse_datetime", "current_time", "timezone_offset",
	"current_timestamp", "current_date", "timestamp_diff", "convert_to_utc",
]



from msb_validation.decorators import (validate_request_data)
from .api_view import (
	ApiView, permissions, api_details,IntraServiceRequestApi,
	TokenUser, SessionData, ApiResponse, RequestInfo, RequestHeaders,
)
from .api_viewset import (ApiViewset, RestResponse)
from .constants import *
from .errors import ErrorViews

__all__ = [
	"ApiView", "ApiViewset", "IntraServiceRequestApi","api_details",
	"validate_request_data", "permissions",
	"TokenUser", "SessionData", "ApiResponse",
	"RequestInfo", "RequestHeaders", "RestResponse"
]



from django.conf import settings
from django.core.serializers import serialize
from django.db import models as Models

from msb_cipher import Cipher
from msb_env.constants import MsbConfigNames as _mcn
from .constants import (COLUMN_NAME_DELETED, COLUMN_NAME_DELETED_BY)
from .msb_model_manager import MsbModelManager


class MsbModel(Models.Model):
	_private_fields: list = []
	_list_field_names: list = []
	_identifier_field: str = ""
	_encrypt_private_fields: bool = getattr(settings, _mcn.MSB_DB_ENCRYPT_PRIVATE_FIELDS, False)
	_pk_is_private_fields: bool = getattr(settings, _mcn.MSB_DB_PK_IS_PRIVATE_FIELD, False)

	class Meta:
		abstract = True

	@property
	def secure_fields(self):
		_secure_fields = self._private_fields if isinstance(self._private_fields, list) else []
		if self._pk_is_private_fields:
			_secure_fields.append(self._meta.pk.attname)
		return _secure_fields

	@property
	def identifier_field_name(self):
		if isinstance(self._identifier_field, str) and len(self._identifier_field) > 0:
			return self._identifier_field
		return ''

	def _get_field_value(self, field_name: str, encrypt: bool = False, default=None):
		try:
			if "." in field_name:
				field_name, relation_name, *_ = field_name.split(".")
				local_field_value = getattr(self, field_name, None)
				field_value = local_field_value._get_field_value(relation_name, default=None)
			else:
				field_value = getattr(self, field_name, default)
				if isinstance(field_value, MsbModel):
					field_value = field_value._get_field_value(field_value.pk_name)

		except Exception as e:
			field_value = None
		finally:

			if (field_name in self.secure_fields) and (encrypt or self._encrypt_private_fields):
				return Cipher.encrypt(field_value)
			return field_value

	@property
	def related_fields(self):
		fields = []
		for field in self._meta.fields:
			if field.get_internal_type() in ['ForeignKey']:
				fields.append(field.name)
		return fields

	@property
	def pk_name(self):
		return self._meta.pk.attname

	@property
	def pk_value(self):
		return getattr(self, self.pk_name) if self.pk_name is not None else ""

	@property
	def identifier(self):
		return f"{getattr(self, self.identifier_field_name)}" if hasattr(self, self.identifier_field_name) else ""

	@property
	def list_field_names(self) -> list:
		return [self.pk_name, *self._list_field_names] if isinstance(self._list_field_names, list) else None

	def dict(self, encrypted: bool = False):
		encrypted = encrypted or self._encrypt_private_fields
		try:
			return {
				k: v if (k not in self.secure_fields or not encrypted) else Cipher.encrypt(v)
				for k, v in super().__dict__.items()
				if not k.startswith('__') and not k.startswith('_') and not callable(k)
			}

		except Exception:
			return dict()

	@property
	def list_fields(self) -> dict:
		if len(self.list_field_names) > 0:
			return {field_name.replace('.', '_'): self._get_field_value(field_name) for field_name in self.list_field_names}
		return None

	@property
	def serialized(self):
		return serialize('python', [self])

	def delete(self, deleted_by=None, using=None, keep_parents=False):
		if hasattr(self, COLUMN_NAME_DELETED):
			setattr(self, COLUMN_NAME_DELETED, True)

		if hasattr(self, COLUMN_NAME_DELETED_BY):
			setattr(self, COLUMN_NAME_DELETED_BY, deleted_by)
		self.save()
		return True

	def recover(self):
		if hasattr(self, COLUMN_NAME_DELETED):
			setattr(self, COLUMN_NAME_DELETED, False)
		self.save()
		return True

	def __str__(self):
		return f"<{self.__class__.__name__} [{self.pk_value}]: {self.identifier}>"

	def __unicode__(self):
		return self.__str__()

	def __repr__(self):
		return self.__str__()

	@property
	def rows(self):
		return self.objects

	objects = MsbModelManager()



from os import environ

from rest_framework.test import APILiveServerTestCase

from .core import (TestConfig, LiveServerThreadWithReuse, const)


class ApiTestResult:
	data = None
	message: str = ''
	code: int = 0
	success: bool = False

	def __init__(self, success=None, code=None, message='', data=None, **kwargs):
		self.data = data if data is not None else []
		self.message = message
		self.code = code
		self.success = success

	@property
	def dict(self):
		return self.__dict__


class ApiTestConfig(TestConfig):

	@property
	def base_url(self):
		return self.get('BASE_URL', default=environ.get('BASE_URL', default='')).rstrip('/')

	def make_url(self, endpoint: str = ''):
		return f"{self.base_url}/{endpoint.replace('.', '/')}"

	@property
	def default_headers(self):
		_headers = {
			const.CONTENT_TYPE_FIELD: const.CONTENT_TYPE_JSON
		}
		if self.user_agent:
			_headers[const.USER_AGENT_FIELD_NAME] = self.user_agent

		return _headers

	@property
	def user_agent(self):
		return self.get(const.CONFIG_USER_AGENT_NAME, default=const.DEFAULT_USER_AGENT)

	@property
	def auth_url(self):
		return self.get(const.CONFIG_AUTH_ENDPOINT, default='')


class ApiTest(APILiveServerTestCase, ApiTestConfig):
	databases = const.DEFAULT_TEST_DATABASES
	port = const.DEFAULT_TEST_SERVER_PORT
	server_thread_class = LiveServerThreadWithReuse

	__auth_tokens: dict = dict()
	_api_endpoints: dict = {}

	def __init__(self, *args, **kwargs):
		APILiveServerTestCase.__init__(self, *args, **kwargs)
		ApiTestConfig.__init__(self)

	def url(self, name: str):
		return self._api_endpoints.get(name)

	def setUp(self) -> None:
		self._authenticate(self.auth_credentials)

	@property
	def access_token(self):
		return self.__auth_tokens.get(const.ACCESS_TOKEN_NAME)

	@property
	def is_authenticated(self):
		return self.access_token not in ['', None]

	@property
	def api_test_result(self):
		return ApiTestResult

	def __log_api_request(self, **kwargs):
		if environ.get('ENVIRONMENT') == 'local':
			logstr = "{sep}Request: {method} {url}\nPayload: {data}\nOptions: {opt}\nResponse: {result}\n{sep}".format(
				sep=f"{'=' * 100}\n", method=kwargs.get('method').upper(), url=kwargs.get('request_url'),
				data=kwargs.get('data'), result=kwargs.get('response'), opt=kwargs.get('opt')
			)
			print(logstr)

	def __make_api_request(self, method='post', endpoint: str = '', data: dict | list = None, _func=None,
	                       **opt) -> ApiTestResult:
		data = dict() if data is None else data
		result = dict()
		try:
			extra_headers = opt.get('headers') if isinstance(opt.get('headers'), dict) else {}
			request_headers = {**self.default_headers, **extra_headers}
			request_url = self.make_url(endpoint=endpoint)

			if self.access_token:
				self.client.credentials(HTTP_AUTHORIZATION=f'Bearer {self.access_token}')

			handler_func = _func if callable(_func) else self.client.post
			api_response = handler_func(request_url, data, 'json', **request_headers)

			if api_response.headers.get(const.CONTENT_TYPE_FIELD) == const.CONTENT_TYPE_JSON:
				result = api_response.json()
			else:
				result = dict(body=api_response.getvalue())
		except Exception as e:
			pass
		finally:
			self.__log_api_request(method=method, request_url=request_url, data=data, opt=opt, response=result)
			result = ApiTestResult(**result)

		# return only the data section of the api response, unless specifically asked for
		return result

	def _authenticate(self, auth_credentials: dict | list = None) -> None:
		if self.is_authenticated:
			return
		auth_result = self.api_post(self.auth_url, payload=auth_credentials)
		if self.assertEquals(list(auth_result.data.keys()), [const.REFRESH_TOKEN_NAME, const.ACCESS_TOKEN_NAME]):
			self.__auth_tokens = dict(auth_result.data)

	def api_get(self, endpoint: str = '', payload: dict | list = None, **opt) -> ApiTestResult:
		return self.__make_api_request('get', endpoint, payload, _func=self.client.get, **opt)

	def api_post(self, endpoint: str = '', payload: dict | list = None, **opt) -> ApiTestResult:
		return self.__make_api_request('post', endpoint, payload, _func=self.client.post, **opt)

	def api_put(self, endpoint: str = '', payload: dict | list = None, **opt) -> ApiTestResult:
		return self.__make_api_request('put', endpoint, payload, _func=self.client.put, **opt)

	def api_delete(self, endpoint: str = '', payload: dict = None, **opt) -> ApiTestResult:
		return self.__make_api_request('delete', endpoint, payload, _func=self.client.delete, **opt)



import logging

from requests.models import Response

from msb_env import (Config)
from .dataclasses import ApiRequestData, ApiResponseWrapper
from .exceptions import ApiRequestExceptions


def make_api_request(request_data: ApiRequestData) -> ApiResponseWrapper:
	from requests import (api, ConnectionError)
	api_response = Response()
	try:
		request_parameters = request_data.get_request_parameters()
		api_response = api.request(**request_parameters)
	except ConnectionError as ce:
		if Config.is_local_env():
			logging.exception(ce)
		raise ApiRequestExceptions.ResourceNotFound
	except Exception as e:
		if Config.is_local_env():
			logging.exception(e)
		raise ApiRequestExceptions.InternalServerError

	return ApiResponseWrapper(api_response)



from msb_exceptions import ApiException


class ApiViewExceptions:
	class UserAuthenticationFailed(ApiException):
		_message = "Failed to authenticate user details."

	class InvalidUserAuthorizationDetails(ApiException):
		_message = "User authorization details not provided."


class ApiViewsetExceptions:
	class SchemaValidationClassNotDefined(ApiException):
		_message = "Api class inherited from 'ApiViewset' must have a 'validation_schema_class' property defined." \
		           "The value of which should be the class which holds your validation rules."



from .api_service import ApiService
from .exceptions import ApiServiceExceptions, CrudApiException
from .isr_service import IntraServiceRequestService

__all__ = ["ApiService", "ApiServiceExceptions", "CrudApiException", "IntraServiceRequestService"]



import logging


class CustomExceptionHandler:

	@staticmethod
	def handle_exception(e):
		return CustomExceptionHandler.handle_error(e)
		logging.exception(e)

	@staticmethod
	def handle_error(e):
		logging.error(f"Error : {e}")



import logging
import os

from msb_env import (NameConst, PathConst)

from .dataclasses import (DjangoMigrationConfig)
from .funcs import (log_to_console, get_django_db_vendor_config)


class DjangoBase:
	def _execute_cmd(self, *args, format=True, log=True):
		from django.core.management import call_command
		if log:
			log_to_console(msg=f"executing {' '.join(args if len(args) > 0 else 'all')} ", format=format)
		call_command(*args)

	def _db_query(self, db_connection, *queries):
		if db_connection and len(queries) > 0:
			with db_connection.cursor() as cursor:
				_result = []
				for query in queries:
					print(f"Executing Query = {query}\n")
					query_result = cursor.execute(query)
					if query.lstrip(' ').lower().startswith(('select', 'show')):
						_result.append(cursor.fetchall())
					else:
						_result.append(query_result)
				return _result[0] if len(_result) == 1 else _result


class DjangoMigration(DjangoBase, DjangoMigrationConfig):

	def run(self):
		try:
			if self.drop_tables_from_db:
				self.__drop_tables_from_database()
			self.__build_migrations()
			self.__migrate_databases()

		except Exception as e:
			logging.exception(e)

	def __init__(self, env: str, **kwargs):
		DjangoMigrationConfig.__init__(self, env=env, **kwargs)

	def __remove_migration_files(self, migration_dir: str):
		if self.is_local_env or self.remove_migration_files:
			log_to_console(msg=f"Removing Migration Files From '{migration_dir}' ")
			for miration_file in os.listdir(migration_dir):
				if miration_file not in ['__init__.py', '__pycache__']:
					os.remove(os.path.join(migration_dir, miration_file))
					log_to_console(miration_file)

	def __drop_tables_from_database(self):
		if not self.is_local_env:
			return logging.warning(msg=f"Dropping Tables only allowed in Local Enviroment")
		from django.db import connections
		for db in self.db_to_drop_tables_from:
			con = connections[db]
			db_config = get_django_db_vendor_config(db_connection=con)
			if db_config:
				log_to_console(f"Dropping Tables From Database[{db}]", format=True)
				table_list = self._db_query(con, db_config.query_to_list_tables)
				queries_to_drop_tables = db_config.queries_to_drop_multiple_tables(table_list, )

				self._db_query(con, *[*queries_to_drop_tables])

	def __build_migrations(self):
		for app_name in self.apps_to_migrate:
			log_to_console(msg=f"Buiding Migration Files For '{app_name}' ", format=True)

			migration_dir = PathConst.APP_DIR_PATH.joinpath(app_name).joinpath(NameConst.MIGRATIONS_DIR_NAME)
			if not os.path.isdir(migration_dir):
				os.mkdir(migration_dir, 0o777)

			self.__remove_migration_files(migration_dir=migration_dir)
			self._execute_cmd("makemigrations", app_name)

	def __migrate_databases(self):
		xargs = ["--no-input"] if not self.is_local_env else []
		for db in set(self.dbs_to_migrate):
			self._execute_cmd("migrate", "--database", db, *xargs)


class DjangoFixtures(DjangoBase):

	def load(self, *from_dirs):
		for fixture_dir in from_dirs:
			fixture_dir = PathConst.FIXTURES_DIR_PATH.joinpath(fixture_dir.lower())
			self.__load_database_fixtures_from_dir(dir=fixture_dir)

	def __init__(self, env: str, file_ext: str = NameConst.YAML_FILE_EXTENTION_NAME):
		self.env = env
		self.file_ext = file_ext

	def __get_list_of_fixtures_to_load_from_dir(self, dir: str):
		if os.path.isdir(dir):
			return sorted(list(filter(lambda f: os.path.splitext(f)[1] == f".{self.file_ext}", os.listdir(dir))))
		else:
			logging.warning(f"Fixture Directory '{dir}' not found.")
			return []

	def __load_database_fixtures_from_dir(self, dir: str):
		_fixtures_list = self.__get_list_of_fixtures_to_load_from_dir(dir=dir)
		_seperator = '-' * 50
		_log = f"## Loading {len(_fixtures_list)} fixtures from '{os.path.basename(dir)}' Dir in following sequence :"
		print(_seperator, _log,)
		for _fixture in _fixtures_list:
			_status = "Success"
			try:
				self._execute_cmd("loaddata", _fixture, format=False, log=False)
			except Exception as e:
				_status = f"Failed->{e}"
			finally:
				print(_fixture,":", _status)



from django.db.models.query import QuerySet

from msb_cipher import Cipher
from msb_dataclasses import Singleton
from msb_dataclasses.search_parameter import SearchParameter
from msb_db import (DEFAULT_QUERY_LIMIT, DEFAULT_QUERY_OFFSET)
from msb_db.models import (MsbModel)
from msb_exceptions import CustomExceptionHandler
from .exceptions import ApiServiceExceptions


class ApiService(CustomExceptionHandler, metaclass=Singleton):
	db_model: MsbModel | None = None
	resource_name: str | None = None
	exception_class_name: ApiServiceExceptions | None = None

	@property
	def resource_name_value(self) -> str | None:
		try:
			if not self.resource_name:
				import re
				class_name = re.split('(?<=.)(?=[A-Z])', self.__class__.__name__.replace("Service", ""))
				return " ".join(class_name)
			return self.resource_name
		except Exception:
			return None

	@property
	def exception_class(self) -> ApiServiceExceptions:
		if not isinstance(self.exception_class_name, ApiServiceExceptions):
			return ApiServiceExceptions
		return self.exception_class_name

	@property
	def cipher(self):
		return Cipher

	def __decrypt(self, data):
		return Cipher.decrypt(data=data)

	def __init__(self):
		if self.db_model is None:
			raise ApiServiceExceptions.InvalidDatabaseModel(resource=self.resource_name_value)

	def search(self, params: SearchParameter) -> QuerySet:
		try:
			model_query = params.get_query(self.db_model)
			return model_query.all()[params.offset:params.offset + params.limit]
		except Exception:
			return QuerySet()

	def create(self, *model_data_list) -> bool:
		try:
			create_status = False
			if len(model_data_list) == 0:
				raise ApiServiceExceptions.InvalidDataForCreateOperation(resource=self.resource_name_value)

			if len(model_data_list) == 1:
				model_data = model_data_list[0]
				_model = self.db_model(**model_data)
				_model.save()
				create_status = True
			else:
				create_status = self.db_model.objects.bulk_create(
					[self.db_model(**model_data) for model_data in model_data_list]
				)
				create_status = True
		except Exception:
			raise ApiServiceExceptions.CreateOperationFailed(resource=self.resource_name_value)
		return create_status

	def retrieve(self, pk=None, silent=False):
		try:
			if pk is None:
				raise ApiServiceExceptions.InvalidPk(resource=self.resource_name_value)
			return self.db_model.objects.get(pk=pk)
		except self.db_model.DoesNotExist:
			if not silent:
				raise ApiServiceExceptions.ResourseDoesNotExists(resource=self.resource_name_value)
		except Exception as e:
			if not silent:
				raise ApiServiceExceptions.RetrieveOperationFailed(resource=self.resource_name_value)
		return None

	def list(self, limit: int = DEFAULT_QUERY_LIMIT, offset: int = DEFAULT_QUERY_OFFSET) -> QuerySet | None:
		try:
			offset = int(offset) if str(offset).isnumeric() else DEFAULT_QUERY_OFFSET
			limit = int(limit) if str(limit).isnumeric() else DEFAULT_QUERY_LIMIT

			fields = [
				i for i in self.db_model._list_field_names
				if (x := getattr(self.db_model, i, None)) and not isinstance(x, property)
			]
			data_set = self.db_model.objects.only(*fields).all()
			return data_set[offset:(limit + offset)] if len(fields) > 0 else None
		except Exception as e:
			raise ApiServiceExceptions.ListOperationFailed(resource=self.resource_name_value)

	def update(self, pk=None, **model_data) -> bool:
		try:
			status = False
			if (pk := self.__decrypt(pk)) is None:
				raise ApiServiceExceptions.InvalidPk(resource=self.resource_name_value)
			model_object = self.db_model.objects.filter(pk=pk)
			if model_object and not (status := model_object.update(**model_data)):
				raise ApiServiceExceptions.UpdateOperationFailed(resource=self.resource_name_value)
			return bool(status)
		except Exception as e:
			raise ApiServiceExceptions.UpdateOperationFailed(resource=self.resource_name_value)

	def delete(self, pk=None) -> bool:
		try:
			status = False

			if (pk := self.__decrypt(pk)) is None:
				raise ApiServiceExceptions.InvalidPk(resource=self.resource_name_value)
			model_object = self.db_model.objects.get(pk=pk)

			if model_object and not (status := model_object.delete()):
				raise ApiServiceExceptions.DeleteOperationFailed(resource=self.resource_name_value)
			return bool(status)
		except Exception as e:
			raise ApiServiceExceptions.DeleteOperationFailed(resource=self.resource_name_value)



from django.db import models

from .constants import (
	COLUMN_NAME_DELETED, COLUMN_NAME_CREATED_AT, COLUMN_NAME_UPDATED_AT,
	COLUMN_NAME_CREATED_BY, COLUMN_NAME_UPDATED_BY
)
from .msb_model import MsbModel


class MsbModelMetaFields(MsbModel):
	class Meta:
		abstract = True

	updated_at = models.DateTimeField(db_column=COLUMN_NAME_UPDATED_AT, auto_now=True)
	updated_by = models.IntegerField(db_column=COLUMN_NAME_UPDATED_BY, null=True, default=None)
	created_at = models.DateTimeField(db_column=COLUMN_NAME_CREATED_AT, auto_now_add=True)
	created_by = models.IntegerField(db_column=COLUMN_NAME_CREATED_BY, null=True, default=None)
	is_deleted = models.BooleanField(db_column=COLUMN_NAME_DELETED, default=False, null=False)



from functools import wraps

from rest_framework import (request as _request, decorators)

from msb_auth.users import TokenUser
from .exceptions import MsbAuthExceptions

require_permission = decorators.permission_classes
require_authentication = decorators.authentication_classes


def management_api(_func):
	@wraps(_func)
	def wrapper(klass, request: _request.Request, **kwargs):
		_user: TokenUser = request.user
		if not _user.has_management_access:
			raise MsbAuthExceptions.UnauthorizedAccess

		return _func(klass, request, **kwargs)

	return wrapper



AUTH_REQUEST_TYPE_FIELD_NAME = "auth_type"
AUTH_REQUEST_USER_FIELD_NAME = "user"
AUTH_REQUEST_PASSWORD_FIELD_NAME = "password"

AUTH_REQUEST_TYPE_SSO = "sso"
AUTH_REQUEST_TYPE_LDAP = "ldap"
AUTH_REQUEST_TYPE_CRED = "cred"
AUTH_REQUEST_TYPE_ISR = "isr-token"

AUTH_TOKEN_ACCESS_TYPE_ISR = "isr"
AUTH_TOKEN_ACCESS_TYPE_USER = "user"


class TokenFieldNames:
	ID = "id"
	USER_ID = "userid"
	AUTH_ID = "auth_userid"
	USERNAME = "username"
	IS_VALID = "is_valid"
	SESSION = "session"
	ACCESS_TYPE = "access_type"
	OWNER = "owner"
	IP = "ip"
	ENV = "env"


MSB_JWT_TOKEN_VALIDATOR = "msb_auth.validators.JwtTokenValidator"
MSB_INTRA_SERVICE_REQUEST_VALIDATOR = "msb_auth.validators.IntraServiceRequestValidator"


class TokenConst:
	username_field = "username"
	userid_field = "id"
	user_email_field = "email"

	access_token_validity_config_name = 'JWT_ACCESS_TOKEN_VALIDITY'
	refresh_token_validity_config_name = 'JWT_REFRESH_TOKEN_VALIDITY'
	signing_key_config_name = 'JWT_TOKEN_SIGNING_KEY'
	verification_key_config_name = 'JWT_TOKEN_VERIFY_KEY'
	token_audiance_config_name = 'JWT_TOKEN_AUDIENCE'
	token_issuer_config_name = 'JWT_TOKEN_ISSUER'


class DefaultJwtConfig:
	username_field = "username"
	userid_claim = "id"
	userid_field = "userid"
	user_email_field = "email"
	user_class = "msb_auth.users.TokenUser"
	auth_rule = 'msb_auth.defaults.jwt_user_auth_rule'
	algorithm_hs256 = 'HS256'
	auth_header_types = ('Bearer',)
	auth_header_name = 'HTTP_AUTHORIZATION'
	token_type_claim = 'token_type'
	jti_claim = 'jti'
	access_token_lifetime = 15
	refresh_token_lifetime = 1440
	rotate_refresh_tokens = False
	blacklist_after_rotation = False
	audience = None
	issuer = None


class MsAuthConst:
	AUTH_VALIDATION_USER_VALUE = "token"
	AUTH_VALIDATION_AUTH_TYPE = "ms_sso"


class LdapAuthConst:
	AUTH_VALIDATION_AUTH_TYPE = "ldap"



import json
from typing import (List, Dict)

from requests import (api, ConnectionError)
from requests.models import Response

from .constants import (HttpHeaders, HttpContentType)
from .dataclasses import (RestRequest, ApiResponseWrapper)
from .exceptions import ApiRequestExceptions

"""
Contains all the parameters required to make a request to the API.
"""


class ApiRequestParameter:

	@property
	def _api_host(self):
		return (self.api_host or "").strip("/")

	@property
	def _api_endpoint(self):
		return (self.endpoint or "").strip("/")

	@property
	def request_is_valid(self) -> bool:
		return True

	@property
	def request_is_json(self) -> bool:
		return self.headers.get(HttpHeaders.CONTENT_TYPE) == HttpContentType.APPLICATION_JSON

	@property
	def request_verify_certificate(self) -> bool:
		return False

	@property
	def request_method(self):
		return self.method

	@property
	def request_headers(self):
		return self.headers

	@property
	def request_query(self):
		return self.query_params

	@property
	def request_url(self):
		_host = self._api_host.strip("/")
		return f"{_host}/{self._api_endpoint}{self.request_query}".strip("/")

	@property
	def request_cookies(self):
		return self.cookies

	@property
	def request_data(self):
		return json.dumps(self.data)

	def get_request_parameters(self):
		_parameters = dict(
			method=self.request_method, url=self.request_url, data=self.request_data,
			verify=self.request_verify_certificate, json=self.request_is_json,
		)

		if self.request_cookies:
			_parameters['cookies'] = self.request_cookies

		if self.request_headers:
			_parameters['headers'] = self.request_headers

		return _parameters

	"""	SETTERS	"""

	def set_endpoint(self, endpoint: str):
		if endpoint:
			self.endpoint = endpoint.rstrip("/")
		return self

	def set_api_host(self, host: str):
		self.api_host = host.rstrip("/") if isinstance(host, str) and len(host) > 0 else None
		return self

	def add_header(self, name: str, value: str):
		self.headers[name] = value
		return self

	def set_data(self, data: [list, dict]):
		self.data = data
		return self

	def set_query_params(self, params: [list, dict]):
		_query_params = ""
		if (params_type := type(params)) in [list, dict]:
			if params_type == dict:
				_query_params = f"""?{"&".join([f"{k}={v}" for k, v in params.items()])}"""
			else:
				_query_params = f"/{'/'.join([str(i) for i in params])}"
		self.query_params = _query_params
		return self

	def set_cookies(self, cookies):
		self.cookies = cookies
		return self

	def set_request_method(self, method: str):
		self.method = method
		return self

	def set_authorization(self, token: str):
		self.add_header(HttpHeaders.AUTHORIZATION, HttpHeaders.BEARER_TOKEN_VALUE.format(token=token))
		return self

	def set_verify_ssl(self, verify: bool):
		self.verify_ssl = verify

	def __init__(self, api_host: str = '', **kwargs):
		self.headers = {HttpHeaders.CONTENT_TYPE: HttpContentType.APPLICATION_JSON}
		self.cookies = None
		self.set_request_method(None)
		self.set_api_host(api_host)
		self.set_endpoint("")
		self.set_query_params([])
		self.set_data({})
		self.set_verify_ssl(kwargs.get('verify_ssl', False))

	def __repr__(self):
		return self.__str__()

	def __str__(self):
		return f"ApiRequestParameter(url={self.request_url}, method={self.request_method},data=" \
		       f"{self.request_data}, headers={self.request_headers}, cookies={self.request_cookies},"


"""
Core Api Request class.
"""


class ApiRequestCore(ApiRequestParameter):

	def _execute(self) -> ApiResponseWrapper:
		api_response = Response()
		try:
			request_parameters = self.get_request_parameters()
			api_response = api.request(**request_parameters)
		except ConnectionError as ce:
			raise ApiRequestExceptions.ResourceNotFound
		except Exception as e:
			raise ApiRequestExceptions.InternalServerError
		finally:
			return ApiResponseWrapper(api_response)

	@property
	def request_is_json(self) -> bool:
		return True

	def get_result(self) -> Dict:
		return self._execute().to_json()

	def __str__(self):
		return f"<{self.__class__.__name__}: {self.request_method.upper()} {self.request_url}>"


class ApiRouterClient(ApiRequestCore):

	def __init__(self, request: RestRequest, **kwargs):
		super(ApiRouterClient, self).__init__(**kwargs)
		if kwargs.get('clone') == True:
			self.set_request_method(request.method)
			self.set_data(request.data)
			self.set_query_params(request.query_params)

		self.headers = request.headers
		self.cookies = request.META.get('HTTP_COOKIE')

	def get_result(self) -> ApiResponseWrapper:
		if not self.request_is_valid:
			raise ApiRequestExceptions.InvalidRequestParameters
		return self._execute()


class ApiRequest(ApiRequestCore):

	def REQUEST(self, method: str, endpoint: str, data: [List | Dict] = None, query_params: [List | Dict] = None, ):
		return self.set_query_params(query_params).set_request_method(method).set_endpoint(endpoint).set_data(data)

	def GET(self, query_params: [List | Dict] = None, endpoint: str = None):
		return self.set_query_params(params=query_params).set_request_method("GET").set_endpoint(endpoint)

	def POST(self, data: [List | Dict] = None, query_params: [List | Dict] = None, endpoint: str = None):
		return self.set_data(data=data).set_query_params(params=query_params).set_request_method("POST").set_endpoint(
			endpoint)

	def PUT(self, data: [List | Dict] = None, query_params: [List | Dict] = None, endpoint: str = None):
		return self.set_data(data=data).set_query_params(params=query_params).set_request_method("PUT").set_endpoint(
			endpoint)

	def DELETE(self, data: [List | Dict] = None, query_params: [List | Dict] = None, endpoint: str = None):
		return self.set_data(data=data).set_query_params(params=query_params).set_request_method("DELETE").set_endpoint(
			endpoint)



from django.utils.functional import cached_property

from .token import AuthToken


class TokenUser(AuthToken):

	@cached_property
	def username(self):
		return self.session.get('username', None)

	@property
	def is_authenticated(self):
		return self.session.get('is_authenticated', False) and self.is_valid

	@cached_property
	def is_admin(self):
		return self.session.get("is_admin") == True

	@property
	def email(self):
		return self.session.get(self._user_email_field)

	def has_access_to(self, path: str):
		return path in self.session.subscriptions



import logging
from argparse import ArgumentParser

from msb_env import NameConst, PathConst

from .constants import (REGEX_TO_SELECT_ENV_VARIABLE, REGEX_TO_REPLACE_ENV_VARIABLE)
from .django import (DjangoMigration, DjangoFixtures)
from .funcs import (require_django, log_to_console)


def build_prod_env_file_from(dev_env_file_path: str = None, prod_env_file_path: str = None):
	import os, re

	dev_env_file_path = dev_env_file_path or PathConst.DEV_ENV_FILE_PATH
	prod_env_file_path = prod_env_file_path or PathConst.PROD_ENV_FILE_PATH

	if os.path.isfile(dev_env_file_path) and os.path.isfile(prod_env_file_path):
		with open(dev_env_file_path, "r+") as dev_env_file:
			dev_env_file_content = dev_env_file.read()

		prod_env_file_content = re.sub(
			REGEX_TO_SELECT_ENV_VARIABLE, REGEX_TO_REPLACE_ENV_VARIABLE, dev_env_file_content
		)

		with open(prod_env_file_path, "w+") as prod_env_file:
			prod_env_file.write(prod_env_file_content)


@require_django
class MsbAppSetupTask:

	@property
	def is_local_env(self):
		from django.conf import settings
		return self.env.lower() == NameConst.LOCAL_ENV_NAME and settings.IS_LOCAL_ENVIRONMENT

	def _initialize_setup_environment(self):
		parser = ArgumentParser(description='Run Msb Setup Script For Django Application')
		parser.add_argument('-env', choices=NameConst.ENV_NAME_LIST, help='Environment Name', required=True)
		args = parser.parse_args()
		self.env = args.env

	def __load_django_db_fixtures(self):
		_fixture_dirs = [NameConst.DEFAULT_FIXTURE_DIR_NAME, *self.fixture_dirs]
		log_to_console(f"Loading Database Fixtures From {_fixture_dirs} Dirs.", format=True)
		database_fixtures = DjangoFixtures(env=self.env)
		database_fixtures.load(*_fixture_dirs)

	def __run_django_db_migrations(self, apps: list, dbs: list, load_fixtures: bool = False, **kwargs):
		try:
			migration = DjangoMigration(env=self.env, **kwargs)
			migration.set_apps_to_migrate(*apps)
			migration.set_dbs_to_migrate(*dbs)
			migration.run()
		except Exception as e:
			logging.exception(e)

	def __init__(self, **kwargs):
		self._initialize_setup_environment()

	def __ask(self, ques: str):
		return str(input(f"{ques}?. : ")).lower().strip(" ")

	def run_database_setup(self, apps: list, dbs: list, fixture_config: dict):

		self.fixture_dirs = fixture_config.get(self.env) or []
		log_to_console(f"Welcome to Database Setup Wizard", format=True)
		print(f"HINT : Reply/Input '1' for yes/agree, anything else to no/skip.\n{'*' * 100}")

		run_db_migrations, drop_tables, load_fixtures = (True, False, True)

		if self.is_local_env:
			if run_db_migrations := bool(self.__ask(f"Run Database Migrations for {apps}")):
				drop_tables = bool(self.__ask("Remove Database tables? This will also remove old migration files."))

			load_fixtures = bool(self.__ask(f"Load Database fixtures for {self.env} Env"))

		if run_db_migrations:
			self.__run_django_db_migrations(
				apps=apps, dbs=dbs, load_fixtures=load_fixtures,
				remove_migration_files=drop_tables,
				drop_tables_from_db=drop_tables,
				db_to_drop_tables_from=[NameConst.DEFAULT_DATABASE_NAME]
			)

		if load_fixtures:
			self.__load_django_db_fixtures()


class MsbAppPreCommitTask:
	pass



from typing import (List, Dict)

from msb_dataclasses import SearchParameterRequest, Singleton
from msb_env import (Config, MsbConfigNames, NameConst)
from .client import (ApiRequest)
from .dataclasses import (MsbApiResponse)


class IntraServiceRequest(ApiRequest):
	"""
	Mdified version of the request class to enable Intra Service requests.
	"""
	_service_name: str = 'IntraServiceRequest'

	def __init__(self, version: int = 1, **kwargs):
		_api_host = "{}/v{}".format(self.msb_intra_service_host, version)
		super(IntraServiceRequest, self).__init__(api_host=_api_host, **kwargs)
		self.set_authorization(token=self.msb_intra_service_request_token)

	@property
	def msb_intra_service_host(self) -> str:
		return Config.get(f"{self._service_name}_URL").as_str(default='')

	@property
	def msb_intra_service_request_token(self) -> str:
		return Config.get(MsbConfigNames.MSB_ISR_TOKEN).as_str(default='')

	def get_result(self) -> MsbApiResponse:
		return self._execute().as_msb_api_response()

	def retrieve(self, endpoint: str, pk: str | int = None) -> MsbApiResponse:
		return self.set_endpoint(endpoint=endpoint).GET(query_params=[pk]).get_result()

	def list(self, endpoint: str, limit: int = -1, offset: int = 0) -> MsbApiResponse:
		return self.set_endpoint(endpoint=endpoint).GET(query_params=dict(limit=limit, offset=offset)).get_result()

	def create(self, endpoint: str, data: [List | Dict] = None) -> MsbApiResponse:
		return self.set_endpoint(endpoint=endpoint).POST(data=data).get_result()

	def search(self, endpoint: str, params: SearchParameterRequest) -> MsbApiResponse:
		return self.set_endpoint(endpoint=f"{endpoint.strip('/')}/search").POST(data=params.__dict__).get_result()

	def update(self, endpoint: str, pk: str | int = None, data: Dict = None) -> MsbApiResponse:
		return self.set_endpoint(endpoint=endpoint).PUT(data=data, query_params=[pk]).get_result()

	def bulk_update(self, endpoint: str, data: List[Dict] = None) -> MsbApiResponse:
		return self.set_endpoint(endpoint=endpoint).PUT(data=data).get_result()

	def delete(self, endpoint: str, pk: str | int = None) -> MsbApiResponse:
		return self.set_endpoint(endpoint=endpoint).DELETE(query_params=[pk]).get_result()

	def bulk_delete(self, endpoint: str, data: List[Dict]) -> MsbApiResponse:
		return self.set_endpoint(endpoint=endpoint).DELETE(data=data).get_result()


class UserServiceApiRequest(IntraServiceRequest, metaclass=Singleton):
	"""
	User Service Api Request Class
	"""

	_service_name = NameConst.USER_SERVICE_NAME


class EmployeeServiceApiRequest(IntraServiceRequest, metaclass=Singleton):
	"""
	Employee Service Api Request Class
	"""
	_service_name = NameConst.EMPLOYEE_SERVICE_NAME


class LeaveServiceApiRequest(IntraServiceRequest, metaclass=Singleton):
	"""
	Leave Service Api Request Class
	"""
	_service_name = NameConst.LEAVE_SERVICE_NAME


class ProjectServiceApiRequest(IntraServiceRequest, metaclass=Singleton):
	"""
	Project Service Api Request Class
	"""

	_service_name = NameConst.PROJECT_SERVICE_NAME


class AutoPilotServiceApiRequest(IntraServiceRequest, metaclass=Singleton):
	"""
	Autopilot Service Api Request Class
	"""
	_service_name = NameConst.AUTOPILOT_SERVICE_NAME


class NotificationServiceApiRequest(IntraServiceRequest, metaclass=Singleton):
	"""
	Notification Service Api Request Class
	"""
	_service_name = NameConst.NOTIFICATION_SERVICE_NAME


class IntraServiceRequestFactory(object):
	"""
	Intra Service Request Factory
	"""

	@staticmethod
	def get_intra_service_request(version: int = 1, **kwargs) -> IntraServiceRequest:
		return IntraServiceRequest(version=version, **kwargs)

	@staticmethod
	def get_user_service_api(version: int = 1, **kwargs) -> UserServiceApiRequest:
		return UserServiceApiRequest(version=version, **kwargs)

	@staticmethod
	def get_employee_service_api(version: int = 1, **kwargs) -> EmployeeServiceApiRequest:
		return EmployeeServiceApiRequest(version=version, **kwargs)

	@staticmethod
	def get_leave_service_api(version: int = 1, **kwargs) -> LeaveServiceApiRequest:
		return LeaveServiceApiRequest(version=version, **kwargs)

	@staticmethod
	def get_project_service_api(version: int = 1, **kwargs) -> ProjectServiceApiRequest:
		return ProjectServiceApiRequest(version=version, **kwargs)

	@staticmethod
	def get_autopilot_service_api(version: int = 1, **kwargs) -> AutoPilotServiceApiRequest:
		return AutoPilotServiceApiRequest(version=version, **kwargs)

	@staticmethod
	def get_notification_service_api(version: int = 1, **kwargs) -> NotificationServiceApiRequest:
		return NotificationServiceApiRequest(version=version, **kwargs)


__all__ = [  # noqa: WPS410
	'IntraServiceRequestFactory',
]



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Union

from requests.models import Response
from rest_framework import request as drf_request
from rest_framework.request import Request as RestRequest
from rest_framework.response import Response as HttpResponse

from msb_dataclasses import Singleton
from msb_env import (Config)
from . import constants as http_const
from .exceptions import ApiRequestExceptions


class HostUrlsConfig(metaclass=Singleton):
	__config_key = "{service_name}_SERVICE_URL"

	def __get_service_host(self, service_name: str):
		_config_name = self.__config_key.format(service_name=service_name.upper())
		return Config.get(_config_name).as_str(default=None)

	def using(self, request_path: str):
		_remote_service_url, *_service_url = (None, [])

		# Separate base service and url from main url
		_service_name, *_service_url = request_path.lstrip("/").split("/")
		_remote_service_url = self.__get_service_host(service_name=_service_name)

		if Config.is_local_env():
			print(f"{_remote_service_url = }")

		return _remote_service_url, "/".join(_service_url)


class ApiRequestData:

	@property
	def _api_host(self):
		return (self.api_host or "").strip("/")

	@property
	def _api_endpoint(self):
		return (self.endpoint or "").strip("/")

	@property
	def request_is_valid(self) -> bool:
		return True

	@property
	def request_is_json(self) -> bool:
		return self.headers.get(http_const.HEADER_NAME_CONTENT_TYPE) == http_const.CONTENT_TYPE_APPLICATION_JSON

	@property
	def request_verify_certificate(self) -> bool:
		return False

	@property
	def request_method(self):
		return self.method

	@property
	def request_headers(self):
		return self.headers

	@property
	def request_query(self):
		return self.query_params

	@property
	def request_url(self):
		_host = self._api_host.strip("/")
		return f"{_host}/{self._api_endpoint}{self.request_query}".strip("/")

	@property
	def request_cookies(self):
		return self.cookies

	@property
	def request_data(self):
		return json.dumps(self.data)

	def get_request_parameters(self):
		_parameters = dict(
			method=self.request_method, url=self.request_url, data=self.request_data,
			verify=self.request_verify_certificate, json=self.request_is_json,
		)

		if self.request_cookies:
			_parameters['cookies'] = self.request_cookies

		if self.request_headers:
			_parameters['headers'] = self.request_headers

		return _parameters

	"""	SETTERS	"""

	def set_endpoint(self, endpoint: str):
		if endpoint:
			self.endpoint = endpoint.rstrip("/")
		return self

	def set_api_host(self, host: str):
		self.api_host = host.rstrip("/") if isinstance(host, str) and len(host) > 0 else None
		return self

	def add_header(self, name: str, value: str):
		self.headers[name] = value
		return self

	def set_data(self, data: [list, dict]):
		self.data = data
		return self

	def set_query_params(self, params: [list, dict]):
		_query_params = ""
		if (params_type := type(params)) in [list, dict]:
			if params_type == dict:
				_query_params = f"""?{"&".join([f"{k}={v}" for k, v in params.items()])}"""
			else:
				_query_params = f"""/{"/".join(params)}"""
		self.query_params = _query_params
		return self

	def set_cookies(self, cookies):
		self.cookies = cookies
		return self

	def set_request_method(self, method: str):
		self.method = method
		return self

	def __init__(self, api_host: str = '', request: RestRequest = None):
		self.method = None
		self.set_api_host(api_host)
		self.set_endpoint("")
		self.set_query_params([])
		self.set_data({})
		self.headers = request.headers if isinstance(request, RestRequest) else dict()
		self.cookies = request.META.get('HTTP_COOKIE') if isinstance(request, RestRequest) else None


class ApiResponseWrapper:
	__response: Response

	@property
	def response(self) -> Response:
		return self.__response

	def __init__(self, response: Response):
		self.__response = response

	def to_json(self) -> dict:
		if self.response.headers.get(http_const.HEADER_NAME_CONTENT_TYPE) == http_const.CONTENT_TYPE_APPLICATION_JSON:
			return self.response.json()

		if self.response.status_code in [404]:
			raise ApiRequestExceptions.ResourceNotFound

	def as_msb_api_response(self) -> MsbApiResponse:
		try:
			return MsbApiResponse(response=self.response)
		except Exception as e:
			return MsbApiResponse(dict())

	def to_http_response(self) -> HttpResponse:
		return HttpResponse(
			data=self.to_json(),
			headers=self.response.headers,
			status=self.response.status_code
		)


class MsbApiResponse:
	__response_body: dict

	def __init__(self, response: Response):
		try:
			self.__response_body = response.json()
		except Exception as e:
			self.__response_body = dict()

	@property
	def code(self):
		return self.__response_body.get("code")

	@property
	def success(self):
		return self.__response_body.get("success") == True

	@property
	def is_valid(self) -> bool:
		return self.success and self.code == 200

	@property
	def data(self):
		_data = self.__response_body.get("data") or []
		return _data

	@property
	def message(self) -> str:
		return self.__response_body.get("message") or "Failure"

	@property
	def body(self) -> str | dict:
		return self.__response_body

	def __str__(self):
		return json.dumps(self.body)


@dataclass
class RequestWrapper:
	request: Union[drf_request.Request] = None

	@property
	def meta(self) -> dict:
		return self.request.META or {}

	@property
	def headers(self) -> dict:
		return self.request.headers or {}

	@property
	def cookie(self):
		return self.meta.get('HTTP_COOKIE')

	@property
	def path(self) -> str:
		return self.meta.get('PATH_INFO')

	@property
	def ip(self):
		return self.headers.get('X-Real-Ip') or self.meta.get('REMOTE_ADDR')

	@property
	def method(self) -> str:
		return self.meta.get('REQUEST_METHOD')

	@property
	def script(self) -> str:
		return self.meta.get('SCRIPT_NAME')

	@property
	def server(self) -> str:
		return self.meta.get('SERVER_NAME')

	@property
	def port(self) -> int:
		return int(self.meta.get('SERVER_PORT'))

	@property
	def protocol(self) -> str:
		return self.meta.get('SERVER_PROTOCOL')

	@property
	def content_type(self) -> str:
		return self.meta.get('CONTENT_TYPE')

	@property
	def query_string(self) -> str:
		return self.meta.get('QUERY_STRING')

	@property
	def authorization(self) -> str:
		return self.meta.get('HTTP_AUTHORIZATION')

	@property
	def user_agent(self) -> str:
		return self.headers.get('User-Agent')

	@property
	def path(self) -> str:
		return self.meta.get('PATH_INFO')
